@@ -3,7 +3,7 @@
 //
 // Code available from: https://verilator.org
 //
-// Copyright 2001-2023 by Wilson Snyder. This program is free software; you
+// Copyright 2001-2024 by Wilson Snyder. This program is free software; you
 // can redistribute it and/or modify it under the terms of either the GNU
 // Lesser General Public License Version 3 or the Perl Artistic License
 // Version 2.0.
@@ -112,7 +112,19 @@ void VerilatedVcd::open(const char* filename) VL_MT_SAFE_EXCLUDES(m_mutex) {
     openNextImp(m_rolloverSize != 0);
     if (!isOpen()) return;
 
-    dumpHeader();
+    printStr("$version Generated by VerilatedVcd $end\n");
+    printStr("$timescale ");
+    printStr(timeResStr().c_str());  // lintok-begin-on-ref
+    printStr(" $end\n");
+
+    // Scope and signal definitions
+    assert(m_indent >= 0);
+    ++m_indent;
+    Super::traceInit();
+    --m_indent;
+    assert(m_indent >= 0);
+
+    printStr("$enddefinitions $end\n\n\n");
 
     // When using rollover, the first chunk contains the header only.
     if (m_rolloverSize) openNextImp(true);
@@ -164,6 +176,7 @@ void VerilatedVcd::openNextImp(bool incFilename) {
         }
     }
     m_isOpen = true;
+    constDump(true);  // First dump must containt the const signals
     fullDump(true);  // First dump must be full
     m_wroteBytes = 0;
 }
@@ -175,49 +188,14 @@ bool VerilatedVcd::preChangeDump() {
 
 void VerilatedVcd::emitTimeChange(uint64_t timeui) {
     printStr("#");
-    printQuad(timeui);
+    const std::string str = std::to_string(timeui);
+    printStr(str.c_str());
     printStr("\n");
 }
 
-void VerilatedVcd::makeNameMap() {
-    // Take signal information from each module and build m_namemapp
-    deleteNameMap();
-    m_namemapp = new NameMap;
-
-    Super::traceInit();
-
-    // Though not speced, it's illegal to generate a vcd with signals
-    // not under any module - it crashes at least two viewers.
-    // If no scope was specified, prefix everything with a "top"
-    // This comes from user instantiations with no name - IE Vtop("").
-    bool nullScope = false;
-    for (const auto& i : *m_namemapp) {
-        const std::string& hiername = i.first;
-        if (!hiername.empty() && hiername[0] == '\t') nullScope = true;
-    }
-    if (nullScope) {
-        NameMap* const newmapp = new NameMap;
-        for (const auto& i : *m_namemapp) {
-            const std::string& hiername = i.first;
-            const std::string& decl = i.second;
-            std::string newname{"top"};
-            if (hiername[0] != '\t') newname += ' ';
-            newname += hiername;
-            newmapp->emplace(newname, decl);
-        }
-        deleteNameMap();
-        m_namemapp = newmapp;
-    }
-}
-
-void VerilatedVcd::deleteNameMap() {
-    if (m_namemapp) VL_DO_CLEAR(delete m_namemapp, m_namemapp = nullptr);
-}
-
 VerilatedVcd::~VerilatedVcd() {
     close();
     if (m_wrBufp) VL_DO_CLEAR(delete[] m_wrBufp, m_wrBufp = nullptr);
-    deleteNameMap();
     if (m_filep && m_fileNewed) VL_DO_CLEAR(delete m_filep, m_filep = nullptr);
     if (parallel()) {
         assert(m_numBuffers == m_freeBuffers.size());
@@ -270,13 +248,6 @@ void VerilatedVcd::printStr(const char* str) {
     }
 }
 
-void VerilatedVcd::printQuad(uint64_t n) {
-    constexpr size_t LEN_STR_QUAD = 40;
-    char buf[LEN_STR_QUAD];
-    VL_SNPRINTF(buf, LEN_STR_QUAD, "%" PRIu64, n);
-    printStr(buf);
-}
-
 void VerilatedVcd::bufferResize(size_t minsize) {
     // minsize is size of largest write.  We buffer at least 8 times as much data,
     // writing when we are 3/4 full (with thus 2*minsize remaining free)
@@ -311,8 +282,7 @@ void VerilatedVcd::bufferFlush() VL_MT_UNSAFE_ONE {
             if (VL_UNCOVERABLE(errno != EAGAIN && errno != EINTR)) {
                 // LCOV_EXCL_START
                 // write failed, presume error (perhaps out of disk space)
-                const std::string msg
-                    = std::string{"VerilatedVcd::bufferFlush: "} + std::strerror(errno);
+                const std::string msg = "VerilatedVcd::bufferFlush: "s + std::strerror(errno);
                 VL_FATAL_MT("", 0, "", msg.c_str());
                 closeErr();
                 break;
@@ -325,141 +295,59 @@ void VerilatedVcd::bufferFlush() VL_MT_UNSAFE_ONE {
     m_writep = m_wrBufp;
 }
 
-//=============================================================================
-// VCD string code
-
-char* VerilatedVcd::writeCode(char* writep, uint32_t code) {
-    *writep++ = static_cast<char>('!' + code % 94);
-    code /= 94;
-    while (code) {
-        --code;
-        *writep++ = static_cast<char>('!' + code % 94);
-        code /= 94;
-    }
-    return writep;
-}
-
 //=============================================================================
 // Definitions
 
 void VerilatedVcd::printIndent(int level_change) {
-    if (level_change < 0) m_modDepth += level_change;
-    assert(m_modDepth >= 0);
-    for (int i = 0; i < m_modDepth; i++) printStr(" ");
-    if (level_change > 0) m_modDepth += level_change;
+    if (level_change < 0) m_indent += level_change;
+    for (int i = 0; i < m_indent; ++i) printStr(" ");
+    if (level_change > 0) m_indent += level_change;
 }
 
-void VerilatedVcd::dumpHeader() {
-    printStr("$version Generated by VerilatedVcd $end\n");
-
-    // Verilator used to put in a $date here.  Although $date is shown in
-    // IEEE examples, and it is common in VCD writers, VCD readers don't
-    // seem to care about it.  Thus, we omit the $date so artifacts are
-    // more likely to be reproducible.  If use cases show up that require
-    // the $date command to be present, it could be re-added with support
-    // for the SOURCE_DATE_EPOCH hook.
-
-    printStr("$timescale ");
-    printStr(timeResStr().c_str());  // lintok-begin-on-ref
-    printStr(" $end\n");
-
-    makeNameMap();
-
-    // Signal header
-    assert(m_modDepth == 0);
-    printIndent(1);
-    printStr("\n");
-
-    // We detect the spaces in module names to determine hierarchy.  This
-    // allows signals to be declared without fixed ordering, which is
-    // required as Verilog signals might be separately declared from
-    // SC module signals.
-
-    // Print the signal names
-    const char* lastName = "";
-    for (const auto& i : *m_namemapp) {
-        const std::string& hiernamestr = i.first;
-        const std::string& decl = i.second;
-
-        // Determine difference between the old and new names
-        const char* const hiername = hiernamestr.c_str();
-        const char* lp = lastName;
-        const char* np = hiername;
-        lastName = hiername;
-
-        // Skip common prefix, it must break at a space or tab
-        for (; *np && (*np == *lp); np++, lp++) {}
-        while (np != hiername && *np && *np != ' ' && *np != '\t') {
-            --np;
-            --lp;
-        }
-        // printf("hier %s\n  lp=%s\n  np=%s\n",hiername,lp,np);
-
-        // Any extra spaces in last name are scope ups we need to do
-        bool first = true;
-        for (; *lp; lp++) {
-            if (*lp == ' ' || (first && *lp != '\t')) {
-                printIndent(-1);
-                printStr("$upscope $end\n");
-            }
-            first = false;
-        }
-
-        // Any new spaces are scope downs we need to do
-        while (*np) {
-            if (*np == ' ') np++;
-            if (*np == '\t') break;  // tab means signal name starts
-            printIndent(1);
-            // Find character after name end
-            const char* sp = np;
-            while (*sp && *sp != ' ' && *sp != '\t' && !(*sp & '\x80')) sp++;
-
-            printStr("$scope ");
-            if (*sp & '\x80') {
-                switch (*sp & 0x7f) {
-                case VLT_TRACE_SCOPE_STRUCT: printStr("struct "); break;
-                case VLT_TRACE_SCOPE_INTERFACE: printStr("interface "); break;
-                case VLT_TRACE_SCOPE_UNION: printStr("union "); break;
-                default: printStr("module ");
-                }
-            } else {
-                printStr("module ");
-            }
-
-            for (; *np && *np != ' ' && *np != '\t'; np++) {
-                if (*np == '[') {
-                    printStr("[");
-                } else if (*np == ']') {
-                    printStr("]");
-                } else if (!(*np & '\x80')) {
-                    *m_writep++ = *np;
-                }
-            }
-            printStr(" $end\n");
-        }
-
-        printIndent(0);
-        printStr(decl.c_str());
+void VerilatedVcd::pushPrefix(const std::string& name, VerilatedTracePrefixType type) {
+    std::string newPrefix = m_prefixStack.back().first + name;
+    switch (type) {
+    case VerilatedTracePrefixType::SCOPE_MODULE:
+    case VerilatedTracePrefixType::SCOPE_INTERFACE:
+    case VerilatedTracePrefixType::STRUCT_PACKED:
+    case VerilatedTracePrefixType::STRUCT_UNPACKED:
+    case VerilatedTracePrefixType::UNION_PACKED: {
+        printIndent(1);
+        printStr("$scope module ");
+        const std::string n = lastWord(newPrefix);
+        printStr(n.c_str());
+        printStr(" $end\n");
+        newPrefix += ' ';
+        break;
     }
+    default: break;
+    }
+    m_prefixStack.emplace_back(newPrefix, type);
+}
 
-    while (m_modDepth > 1) {
+void VerilatedVcd::popPrefix() {
+    switch (m_prefixStack.back().second) {
+    case VerilatedTracePrefixType::SCOPE_MODULE:
+    case VerilatedTracePrefixType::SCOPE_INTERFACE:
+    case VerilatedTracePrefixType::STRUCT_PACKED:
+    case VerilatedTracePrefixType::STRUCT_UNPACKED:
+    case VerilatedTracePrefixType::UNION_PACKED:
         printIndent(-1);
         printStr("$upscope $end\n");
+        break;
+    default: break;
     }
-
-    printIndent(-1);
-    printStr("$enddefinitions $end\n\n\n");
-    assert(m_modDepth == 0);
-
-    // Reclaim storage
-    deleteNameMap();
+    m_prefixStack.pop_back();
+    assert(!m_prefixStack.empty());
 }
 
 void VerilatedVcd::declare(uint32_t code, const char* name, const char* wirep, bool array,
-                           int arraynum, bool tri, bool bussed, int msb, int lsb) {
+                           int arraynum, bool bussed, int msb, int lsb) {
     const int bits = ((msb > lsb) ? (msb - lsb) : (lsb - msb)) + 1;
 
-    const bool enabled = Super::declCode(code, name, bits, tri);
+    const std::string hierarchicalName = m_prefixStack.back().first + name;
+
+    const bool enabled = Super::declCode(code, hierarchicalName, bits);
 
     if (m_suffixes.size() <= nextCode() * VL_TRACE_SUFFIX_ENTRY_SIZE) {
         m_suffixes.resize(nextCode() * VL_TRACE_SUFFIX_ENTRY_SIZE * 2, 0);
@@ -472,93 +360,95 @@ void VerilatedVcd::declare(uint32_t code, const char* name, const char* wirep, b
 
     if (!enabled) return;
 
-    // Split name into basename
-    // Spaces and tabs aren't legal in VCD signal names, so:
-    // Space separates each level of scope
-    // Tab separates final scope from signal name
-    // Tab sorts before spaces, so signals nicely will print before scopes
-    // Note the hiername may be nothing, if so we'll add "\t{name}"
-    std::string nameasstr = namePrefix() + name;
-    std::string hiername;
-    std::string basename;
-    for (const char* cp = nameasstr.c_str(); *cp; cp++) {
-        if (isScopeEscape(*cp)) {
-            // Ahh, we've just read a scope, not a basename
-            if (!hiername.empty()) hiername += " ";
-            hiername += basename;
-            basename = "";
-        } else {
-            basename += *cp;
-        }
+    // Create the VCD code and build the suffix array entry
+    char vcdCode[VL_TRACE_SUFFIX_ENTRY_SIZE];
+    {
+        // Render the VCD code
+        char* vcdCodeWritep = vcdCode;
+        uint32_t codeEnc = code;
+        do {
+            *vcdCodeWritep++ = static_cast<char>('!' + codeEnc % 94);
+            codeEnc /= 94;
+        } while (codeEnc--);
+        *vcdCodeWritep = '\0';
+        const size_t vcdCodeLength = vcdCodeWritep - vcdCode;
+        assert(vcdCodeLength <= VL_TRACE_MAX_VCD_CODE_SIZE);
+        // Build suffix array entry
+        char* const entryBeginp = &m_suffixes[code * VL_TRACE_SUFFIX_ENTRY_SIZE];
+        entryBeginp[0] = ' ';  // Separator
+        // 1 bit values don't have a ' ' separator between value and string code
+        char* entryWritep = bits == 1 ? entryBeginp : entryBeginp + 1;
+        // Use memcpy as we know the size, and strcpy is flagged unsafe
+        std::memcpy(entryWritep, vcdCode, vcdCodeLength);
+        entryWritep += vcdCodeLength;
+        // Line terminator
+        *entryWritep++ = '\n';
+        // Set length of suffix (used to increment write pointer)
+        assert(entryWritep <= entryBeginp + VL_TRACE_SUFFIX_ENTRY_SIZE - 1);
+        entryBeginp[VL_TRACE_SUFFIX_ENTRY_SIZE - 1] = static_cast<char>(entryWritep - entryBeginp);
     }
-    hiername += "\t" + basename;
 
-    // Print reference
+    // Assemble the declaration
     std::string decl = "$var ";
-    decl += wirep;  // usually "wire"
-
-    constexpr size_t bufsize = 1000;
-    char buf[bufsize];
-    VL_SNPRINTF(buf, bufsize, " %2d ", bits);
-    decl += buf;
-    // Add string code to decl
-    char* const endp = writeCode(buf, code);
-    *endp = '\0';
-    decl += buf;
-    // Build suffix array entry
-    char* const entryp = &m_suffixes[code * VL_TRACE_SUFFIX_ENTRY_SIZE];
-    const size_t length = endp - buf;
-    assert(length <= VL_TRACE_MAX_VCD_CODE_SIZE);
-    // 1 bit values don't have a ' ' separator between value and string code
-    const bool isBit = bits == 1;
-    entryp[0] = ' ';  // Separator
-    // Use memcpy as we checked size above, and strcpy is flagged unsafe
-    std::memcpy(entryp + !isBit, buf,
-                std::strlen(buf));  // Code (overwrite separator if isBit)
-    entryp[length + !isBit] = '\n';  // Replace '\0' with line termination '\n'
-    // Set length of suffix (used to increment write pointer)
-    entryp[VL_TRACE_SUFFIX_ENTRY_SIZE - 1] = static_cast<char>(length + !isBit + 1);
-    decl += " ";
-    decl += basename;
+    decl += wirep;
+    decl += ' ';
+    decl += std::to_string(bits);
+    decl += ' ';
+    decl += vcdCode;
+    decl += ' ';
+    decl += lastWord(hierarchicalName);
     if (array) {
-        VL_SNPRINTF(buf, bufsize, "[%d]", arraynum);
-        decl += buf;
-        hiername += buf;
+        decl += '[';
+        decl += std::to_string(arraynum);
+        decl += ']';
     }
     if (bussed) {
-        VL_SNPRINTF(buf, bufsize, " [%d:%d]", msb, lsb);
-        decl += buf;
+        decl += " [";
+        decl += std::to_string(msb);
+        decl += ':';
+        decl += std::to_string(lsb);
+        decl += ']';
     }
     decl += " $end\n";
-    m_namemapp->emplace(hiername, decl);
+    printIndent(0);
+    printStr(decl.c_str());
 }
 
-void VerilatedVcd::declEvent(uint32_t code, const char* name, bool array, int arraynum) {
-    declare(code, name, "event", array, arraynum, false, false, 0, 0);
+void VerilatedVcd::declEvent(uint32_t code, uint32_t fidx, const char* name, int dtypenum,
+                             VerilatedTraceSigDirection, VerilatedTraceSigKind,
+                             VerilatedTraceSigType, bool array, int arraynum) {
+    declare(code, name, "event", array, arraynum, false, 0, 0);
 }
-void VerilatedVcd::declBit(uint32_t code, const char* name, bool array, int arraynum) {
-    declare(code, name, "wire", array, arraynum, false, false, 0, 0);
+void VerilatedVcd::declBit(uint32_t code, uint32_t fidx, const char* name, int dtypenum,
+                           VerilatedTraceSigDirection, VerilatedTraceSigKind,
+                           VerilatedTraceSigType, bool array, int arraynum) {
+    declare(code, name, "wire", array, arraynum, false, 0, 0);
 }
-void VerilatedVcd::declBus(uint32_t code, const char* name, bool array, int arraynum, int msb,
-                           int lsb) {
-    declare(code, name, "wire", array, arraynum, false, true, msb, lsb);
+void VerilatedVcd::declBus(uint32_t code, uint32_t fidx, const char* name, int dtypenum,
+                           VerilatedTraceSigDirection, VerilatedTraceSigKind,
+                           VerilatedTraceSigType, bool array, int arraynum, int msb, int lsb) {
+    declare(code, name, "wire", array, arraynum, true, msb, lsb);
 }
-void VerilatedVcd::declQuad(uint32_t code, const char* name, bool array, int arraynum, int msb,
-                            int lsb) {
-    declare(code, name, "wire", array, arraynum, false, true, msb, lsb);
+void VerilatedVcd::declQuad(uint32_t code, uint32_t fidx, const char* name, int dtypenum,
+                            VerilatedTraceSigDirection, VerilatedTraceSigKind,
+                            VerilatedTraceSigType, bool array, int arraynum, int msb, int lsb) {
+    declare(code, name, "wire", array, arraynum, true, msb, lsb);
 }
-void VerilatedVcd::declArray(uint32_t code, const char* name, bool array, int arraynum, int msb,
-                             int lsb) {
-    declare(code, name, "wire", array, arraynum, false, true, msb, lsb);
+void VerilatedVcd::declArray(uint32_t code, uint32_t fidx, const char* name, int dtypenum,
+                             VerilatedTraceSigDirection, VerilatedTraceSigKind,
+                             VerilatedTraceSigType, bool array, int arraynum, int msb, int lsb) {
+    declare(code, name, "wire", array, arraynum, true, msb, lsb);
 }
-void VerilatedVcd::declDouble(uint32_t code, const char* name, bool array, int arraynum) {
-    declare(code, name, "real", array, arraynum, false, false, 63, 0);
+void VerilatedVcd::declDouble(uint32_t code, uint32_t fidx, const char* name, int dtypenum,
+                              VerilatedTraceSigDirection, VerilatedTraceSigKind,
+                              VerilatedTraceSigType, bool array, int arraynum) {
+    declare(code, name, "real", array, arraynum, false, 63, 0);
 }
 
 //=============================================================================
 // Get/commit trace buffer
 
-VerilatedVcd::Buffer* VerilatedVcd::getTraceBuffer() {
+VerilatedVcd::Buffer* VerilatedVcd::getTraceBuffer(uint32_t fidx) {
     VerilatedVcd::Buffer* const bufp = new Buffer{*this};
     if (parallel()) {
         // Note: This is called from VerilatedVcd::dump, which already holds the lock
@@ -683,8 +573,8 @@ void VerilatedVcdBuffer::finishLine(uint32_t code, char* writep) {
 // so always inline them.
 
 VL_ATTR_ALWINLINE
-void VerilatedVcdBuffer::emitEvent(uint32_t code, VlEvent newval) {
-    const bool triggered = newval.isTriggered();
+void VerilatedVcdBuffer::emitEvent(uint32_t code, const VlEventBase* newval) {
+    const bool triggered = newval->isTriggered();
     // TODO : It seems that untriggered events are not filtered
     // should be tested before this last step
     if (triggered) {
